//===--- WorkloadManager.h - Geant-V ----------------------------*- C++ -*-===//
//
//                     Geant-V Prototype
//
//===----------------------------------------------------------------------===//
/**
 * @file WorkloadManager.h
 * @brief Definition of workload manager in Geant-V prototype
 * @author Andrei Gheata
 */
//===----------------------------------------------------------------------===//

#ifndef GEANT_WORKLOADMANAGER
#define GEANT_WORKLOADMANAGER

#include <thread>
#include <list>
#include "priority_queue.h"
#include "dcqueue.h"
#include "condition_locker.h"

#ifndef USE_VECGEOM_NAVIGATOR
#include "TGeoManager.h"
#endif

#include "GeantTrack.h"
#include "GeantPropagator.h"

#ifdef USE_ROOT
#include "TThreadMergingServer.h"
#endif

class GeantBasketMgr;
class GeantBasket;
class GeantScheduler;
class TaskBroker;

/**
 * @brief WorkloadManager class
 * @details Main work manager class. This creates and manages all the worker threads,
 * has pointers to the synchronization objects, but also to the currently
 * transported baskets.
 */
class WorkloadManager {
protected:
  int fNthreads;                                     /** Number of managed threads */
  int fNbaskets;                                     /** Total number of baskets */
  int fBasketGeneration;                             /** Basket generation */
  int fNbasketgen;                                   /** Number of baskets to transport in the current generation */
  int fNidle;                                        /** Number of idle workers */
  int fNminThreshold;                                /** Minimum number of tracks in a basket to trigger transport */
  int fNqueued;                                      /** Number of chunks queued */
  int *fBtogo;                                       /** Array of baskets to be processed in the next generation */
  int fSchId;                                        /** Thread id for the scheduler */
  bool fStarted;                                       /** Start flag */
  bool fStopped;                                       /** Stop flag */
  Geant::priority_queue<GeantBasket *> *fFeederQ;      /** Queue of transportable baskets */
  Geant::priority_queue<GeantBasket *> *fTransportedQ; /** Queue of transported baskets */
  Geant::priority_queue<GeantBasket *> *fDoneQ;        /** Thread "all work done" queue */
  static WorkloadManager *fgInstance;                  /** Singleton instance */
  std::vector<std::thread> fListThreads;               /** Vector of threads */
  bool fFlushed;                                       /** Buffer flushed */
  bool fFilling;                                       /** Worker queue is filling */
  int fMonQueue;                                     /** Monitor the work queue */
  int fMonMemory;                                    /** Monitor the memory */
  int fMonBasketsPerVol;                             /** Monitor baskets per volume */
  int fMonVectors;                                   /** Monitor vector scheduling */
  int fMonConcurrency;                               /** Monitor concurrency */
  int fMonTracksPerEvent;                            /** Monitor tracks status per event */
  int fMonTracks;                                    /** Monitor number of tracks */
  int fMaxThreads;                                   /** Maximum number of threads */
  GeantScheduler *fScheduler;                          /** Main basket scheduler */

  TaskBroker *fBroker;         /** Pointer to the coprocessor broker, this could be made a collection. */
  int *fWaiting;             /** ![fNthreads+1] Threads in waiting flag */
  condition_locker fSchLocker; /** Scheduler locker */
  condition_locker fGbcLocker; /** Garbage collector locker */
  int fLastEvent;            /** Last transported event */

  #ifdef USE_ROOT
  dcqueue<TBufferFile*>* fOutputIO;            /** Queue of buffers to be merged for IO **/ 
  Geant::TThreadMergingServer* fMergingServer;
  #endif
  /**
   * @brief WorkloadManager parameterized constructor
   *
   * @param  nthreads Number of threads foe workload manager
   */
  WorkloadManager(int nthreads);

public:
  /** @brief WorkloadManager destructor */
  virtual ~WorkloadManager();

  /** @brief Create basket function */
  void CreateBaskets();

  enum class FeederResult : char { kNone, kFeederWork, kStopProcessing };

  /** @brief Call Feeder (if needed) and check exit condition. */
  FeederResult CheckFeederAndExit(GeantBasketMgr &prioritizer, GeantPropagator &propagator, Geant::GeantTaskData &td);

  /** @brief Function for feeder queue of transportable baskets */
  Geant::priority_queue<GeantBasket *> *FeederQueue() const { return fFeederQ; }

  /**
   * @brief Function that provides trasported queue for baskets
   */
  Geant::priority_queue<GeantBasket *> *TransportedQueue() const { return fTransportedQ; }

  /**
   * @brief Function that provides thread's "all work done" queue
   */
  Geant::priority_queue<GeantBasket *> *DoneQueue() const { return fDoneQ; }
  //   GeantObjectPool<VolumePath_t>
  //   rr_pool<VolumePath_t>
  //                      *NavStates() const   {return fNavStates;}

  /**
   * @brief Function that provides IO queue 
   */
  #ifdef USE_ROOT
  dcqueue<TBufferFile*>*  IOQueue() const { return fOutputIO; }
  Geant::TThreadMergingServer* MergingServer() const { return fMergingServer; }
  #endif
  /** @brief Function that returns number of managed threads */
  int GetNthreads() const { return fNthreads; }

  /** @brief Function that returns total number of baskets */
  int GetNbaskets() const { return fNbaskets; }

  /** @brief Function that returns threads in waiting array */
  int *GetWaiting() const { return fWaiting; }

  /** @brief Function that returns number of waiting threads */
  int GetNwaiting() const {
    int nwaiting = 0;
    for (int i = 0; i < fNthreads; ++i)
      nwaiting += fWaiting[i];
    return nwaiting;
  }

  /** @brief Function that returns number of threads actually working */
  int GetNworking() const { return (fNthreads - GetNwaiting()); }

  /** @brief Function returning the number of monitored features */
  int GetMonFeatures() const;

  /** @brief Check if a monitoring feature is enabled */
  bool IsMonitored(GeantPropagator::EGeantMonitoringType feature) const;

  /** @brief Enable monitoring a feature */
  void SetMonitored(GeantPropagator::EGeantMonitoringType feature, bool flag = true);

  /** @brief Function that returns main basket scheduler */
  GeantScheduler *GetScheduler() const { return fScheduler; }

  /** @brief Get the scheduler thread id */
  int GetSchId() const { return fSchId; }

  /** @brief Set scheduler thread id */
  void SetSchId(int id) { fSchId = id; }

  /** @brief Function that returns scheduler locker */
  condition_locker &GetSchLocker() { return fSchLocker; }

  /** @brief Function that returns garbage collector locker */
  condition_locker &GetGbcLocker() { return fGbcLocker; }

  /**
   * @brief Function that create workload manager instance
   *
   * @param nthreads Number of threads (by default 0)
   */
  static WorkloadManager *Instance(int nthreads = 0);

  /** @brief Function that check if buffer is flushed */
  bool IsFlushed() const { return fFlushed; }

  /** @brief Function that check if worker queue is filling */
  bool IsFilling() const { return fFilling; }

  /** @brief Function that check stop flag */
  bool IsStopped() const { return fStopped; }

  /** @brief Getter for last transported event */
  int LastEvent() const { return fLastEvent; }

  /** @brief Setter for last transported event */
  void SetLastEvent(int n) { fLastEvent = n; }

  /** @brief Function that provide stop process by setting Stop flag = True */
  void Stop() { fStopped = true; }

  /** @brief Setter for buffer flushing */
  void SetFlushed(bool flag) { fFlushed = flag; }

  /** @brief Function that returns basket generation */
  int GetBasketGeneration() const { return fBasketGeneration; }

  /** @brief Print function */
  void Print(const char *option = "") const;

  /** @brief  Setter for task broker */
  void SetTaskBroker(TaskBroker *broker);

  /** @brief  Setter for task broker */
  TaskBroker *GetTaskBroker() { return fBroker; }

#if USE_VECGEOM_NAVIGATOR == 1
  /**
   * @brief Tell the task broker(s) to load the geometry.
   *
   * @param Volume to load
   */
  bool LoadGeometry(vecgeom::VPlacedVolume const *const volume = nullptr);
#endif

  void SetMaxThreads(int nthreads) {
    fMaxThreads = nthreads;
#ifndef USE_VECGEOM_NAVIGATOR
    gGeoManager->SetMaxThreads(nthreads);
#endif
  }

  int ThreadId();

  /** @brief Getter for the global transport threshold */
  int GetNminThreshold() const { return fNminThreshold; }

  /** @brief Setter for the global transport threshold */
  void SetNminThreshold(int thr) { fNminThreshold = thr; }

  /** @brief Function that provides start process of threads  */
  void StartThreads();

  /** @brief Joins all threads at the end of processing */
  void JoinThreads();

  /** @brief Thread function for the main scheduler */
  static void *MainScheduler(void *arg);

  /**
   * @brief Function that provides garbage collector thread
   *
   * @param arg Arguments to be passed in the function
   */
  static void *GarbageCollectorThread();

  /**
   * @brief Function that provides monitoring thread
   *
   * @param arg Arguments to be passed in the function
   */
  static void *MonitoringThread();

  /**
   * @brief Function that provides output thread
   *
   * @param arg Arguments to be passed in the function
   */
  
  static void *OutputThread();

  /**
   * @brief Function that provides transporting tracks
   *
   * @param arg Arguments to be passed in the function
   */
  static void *TransportTracks();

  /**
   * @brief Function that provides transport tracks in coprocessor
   *
   * @param arg Arguments to be passed in the function
   */
  static void *TransportTracksCoprocessor(TaskBroker *broker);

  /** @brief Function that provides waiting of workers */
  void WaitWorkers();

private:
  /**
   * @brief Copy constructor for WorkloadManager
   * @details Still not implemented
   */
  WorkloadManager(const WorkloadManager &);

  /**
   * @brief Operator =
   * @todo Still not implemented
   */
  WorkloadManager &operator=(const WorkloadManager &);
};
#endif
